import json
import os

KEY_FILE = "~/.uh-keys"


def prompt_key(service: str) -> str:
    """
    Prompt the user for a key for the given service, returning it as a string
    """
    key = input(f">>> Need a key for {service}, paste it here: ").strip()
    return key


def load_keys(key_file: str = KEY_FILE) -> dict[str, str]:
    """
    Load the keys from disk and return them
    """
    key_file = os.path.expanduser(key_file)

    if os.path.exists(key_file):
        with open(key_file, "rt") as f:
            return json.load(f)
    else:
        return {}


def save_keys(keys: dict, key_file: str = KEY_FILE):
    """
    Save all the keys to disk
    """
    key_file = os.path.expanduser(key_file)
    with open(key_file, "wt") as f:
        json.dump(keys, f, indent=4)

    os.chmod(key_file, 0o600)


def save_key(service: str, key: str, key_file: str = KEY_FILE):
    """
    Save the service's key to disk
    """
    keys = load_keys(key_file)

    if key:
        print(f">>> Saving {service} key to {key_file}")
        keys[service] = key
    else:
        print(f">>> Removing {service} key from {key_file}")
        keys.pop(service, None)

    save_keys(keys, key_file)


def load_key(service: str, key_file: str = KEY_FILE) -> str | None:
    """
    Load the key from disk, returning it.
    """
    keys = load_keys(key_file)

    return keys.get(service, None)


def get_key(service: str) -> str:
    """
    Loads the service's key from ~/.uh-keys, or prompts for it and saves it there.
    """
    key = load_key(service)
    if not key:
        key = prompt_key(service)
        save_key(service, key)

    return key



"""
For calling the OpenAI completion model.
"""

import openai

from .util import clean


def query(model: str, system_prompt: str, user_prompt: str, key: str) -> str:
    """
    Calls the OpenAI API with the query from the command line.
    """
    openai.api_key = key

    try:
        response = openai.ChatCompletion.create(
            model=model,
            messages=[
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": user_prompt},
            ],
        )
    except openai.error.AuthenticationError:
        raise Exception(
            "OpenAI key doesn't work. You might need to delete it from ~/.uh-keys"
        )

    response = response["choices"][0]["message"]["content"]
    return clean(response)



"""
Contains main entrypoint
"""

import importlib
import os
import sys
from typing import Callable

from .config import CONFIG_FILE, get_config
from .keys import get_key
from .vars import apply_vars, get_vars


def show_help():
    """
    uh
        command line help on the command line

    Usage: uh [query]

    Example:
        user@host$ uh get the time
        date +%H:%M:%S

    config is in ~/.uh_config. Each field is parsed with mustache.
    Access tokens go in ~/.uh-keys
    """
    doc = show_help.__doc__
    halp = "\n".join((line[4:] for line in doc.split("\n")))

    print(halp)


def debug_print_query(**kwargs):
    return "failed query attempt: " + str(kwargs)


def get_func(module_name: str, func_name="query") -> Callable:
    """
    Returns a function in the given module name.
    """
    old_path = sys.path.copy()
    pwd = os.getcwd()
    sys.path = [p for p in sys.path if p not in ("", ".", pwd)]

    try:
        module = importlib.import_module(module_name)
        sys.path = old_path
    except ModuleNotFoundError:
        print(f"Module {module_name} not found. Using debug printer instead.")
        return debug_print_query

    return getattr(module, func_name)


def main() -> int:
    """
    Entrypoint for the CLI.
    """
    config = get_config()
    current = config.get("current", None)

    needs_help = len(sys.argv) == 1 or (
        len(sys.argv) == 2 and any(halp in sys.argv[1] for halp in ("halp", "help"))
    )

    if needs_help:
        show_help()
        return 0

    if current not in config["services"]:
        choices = ", ".join(config["services"].keys())
        print(f"'{current}' is not one of: {choices}")
        print(f"edit {CONFIG_FILE}")
        return 0

    current_config = config["services"][current]
    vars = get_vars()

    key = None

    if current_config.get("use_key", False):
        key = get_key(current)
        if not key:
            print(f"{current} needs an access key.")
            return 1

    vars["key"] = key

    params = apply_vars(vars, current_config["params"])

    query = get_func(current_config["module"])

    response = query(**params)
    print(response)


if __name__ == "__main__":
    sys.exit(main())



"""
Gets template variables that can be inserted into the context.
"""

import os
import platform
import subprocess
import sys
from typing import Any


def get_os() -> str:
    """
    Returns a string describing the OS.
    """
    system = platform.system()
    if system == "Linux":
        try:
            return subprocess.check_output(["lsb_release", "-ds"]).decode().strip()
        except Exception:
            with open("/etc/os-release") as f:
                for line in f:
                    if line.startswith("PRETTY_NAME="):
                        return line.strip().split("=")[1].strip('"')
    elif system == "Darwin":
        lines = subprocess.check_output(["sw_vers"]).decode().split("\n")
        words = [word.split(":", maxsplit=1)[-1].strip() for word in lines if word]
        return " ".join(words)
    elif system == "Windows":
        return f"Windows {platform.release()}"
    elif "BSD" in system:
        release = subprocess.check_output(["uname", "-r"]).decode().strip()
        return f"{system} {release}"

    return system


def get_shell() -> str:
    """
    Returns the user's shell executable name.
    """
    shell_cmd = os.environ.get("SHELL", os.environ.get("ComSpec", "unknown"))
    shell_cmd = shell_cmd.replace("\\", "/")
    return os.path.basename(shell_cmd)


def apply_vars(vars: dict, template: Any) -> Any:
    """
    Replaces strings in a json-style object tree. Use {var}
    """
    if isinstance(template, str):
        return template.format(**vars)
    elif isinstance(template, dict):
        return {
            apply_vars(vars, key): apply_vars(vars, value)
            for key, value in template.items()
        }
    elif isinstance(template, list):
        return [apply_vars(vars, item) for item in template]
    else:
        return template


def get_vars() -> dict:
    """
    Gets variables that can be replaced with templates.
    To use one, reference it like {var} in any string field.

    Current keys are:

    - shell: The user's shell executable name.
    - os: A string describing the OS and its version.
    - query: The query string.
    - pwd: The current working directory.
    - key: the secret key for the service (blank here, requested interactively if needed)
    """
    return {
        "shell": get_shell(),
        "os": get_os(),
        "query": " ".join(sys.argv[1:]),
        "key": None,  # updated later if required
        # I'm not sending this one, you can if you want.
        "pwd": os.getcwd(),
    }



"""
Generic web API functions
"""

import json

import requests


def query(
    endpoint: str,
    method: str,
    query: str,
    post_data: str,
    headers: dict[str, str],
    response_path: list[str | int],
) -> str:
    """
    Standard web API request.
    """

    if isinstance(post_data, dict):
        post_data = json.dumps(post_data)

    response = requests.request(
        method, endpoint, params=query, data=post_data, headers=headers
    )

    return extract(response, response_path)


def extract(response: requests.Response, response_path: list[str | int]) -> str:
    """
    If there's a response_path list, treat the response as JSON and the list as
    the path to the key we want. Use strings for property names, ints for list
    indices, and you can use negative indices to reference the end of a list
    (i.e. -1 for the last item.)
    """
    if not response_path:
        return response.text

    current = response.json()
    for p in response_path:
        current = current[p]
    return current



def clean(response: str) -> str:
    """
    Cleans the response from the API, because we can't trust it to play nicely.
    """
    split_response = response.split("```")
    if len(split_response) == 3:
        response = split_response[1]

    split_response = [line for line in response.split("\n") if line.strip()]
    response = split_response[0].strip() if split_response else ""

    if not response:
        return "uh, dunno. sorry :("

    return response



import json
import os
import shutil

CONFIG_FILE = "~/.uh-config"


def reset_config(path: str = CONFIG_FILE):
    """
    Resets the config file to the template
    """
    dest = os.path.expanduser(path)

    if os.path.exists(dest):
        os.remove(dest)

    this_file = os.path.realpath(__file__)
    this_dir = os.path.dirname(this_file)
    source = f"{this_dir}/default_config.json"

    shutil.copyfile(source, dest)


def save_config(config, path: str = CONFIG_FILE):
    """
    Saves the config file
    """
    config_file = os.path.expanduser(path)
    with open(config_file, "wt") as f:
        json.dump(config, f, indent=4)


def get_config(path: str = CONFIG_FILE):
    """
    Reads the config file, creating it from template if it doesn't exist
    """
    config_file = os.path.expanduser(path)
    if not os.path.exists(config_file):
        reset_config(path)

    with open(config_file) as f:
        config = json.load(f)

    return config
